@@ -9,38 +9,37 @@ SPDX-License-Identifier: GPL-3.0-only
 # TODO: Notifying locks on profile and firmware upload
 
 import asyncio
-import atexit
+import hashlib
 import json
 import logging
 import time
 
-from copy import copy
+from copy import copy, deepcopy
 from typing import Union, Dict, Coroutine, Optional, List, Callable
 
-from bleak import BleakClient
-from bleak.backends.bluezdbus.client import BleakClientBlueZDBus
 from bleak.backends.device import BLEDevice
 from bleak.backends.scanner import AdvertisementData
 
 from pyDE1.bleak_client_wrapper import BleakClientWrapped
 
-import pyDE1.default_logger
-
-from pyDE1.de1.exceptions import *
+from pyDE1.exceptions import *
 from pyDE1.de1.ble import UnsupportedBLEActionError
 
+import pyDE1.database.insert as db_insert
+import aiosqlite
+
 # general utilities
 
 from pyDE1.de1.c_api import \
     PackedAttr, RequestedState, ReadFromMMR, WriteToMMR, StateInfo, \
     FWMapRequest, FWErrorMapRequest, FWErrorMapResponse, \
     API_MachineStates, API_Substates, MAX_FRAMES, get_cuuid, MMR0x80LowAddr, \
-    packed_attr_from_cuuid, pack_one_mmr0x80_write
+    packed_attr_from_cuuid, pack_one_mmr0x80_write, MMRGHCInfoBitMask
 
 from pyDE1.de1.ble import CUUID
 from pyDE1.de1.notifications import NotificationState, MMR0x80Data
-from pyDE1.de1.profile import ProfileByFrames, DE1ProfileValidationError
-
+from pyDE1.de1.profile import Profile, ProfileByFrames, \
+    DE1ProfileValidationError
 
 from pyDE1.i_target_setter import I_TargetSetter
 from pyDE1.event_manager.events import ConnectivityState, ConnectivityChange
@@ -55,27 +54,24 @@ import pyDE1.de1.handlers
 from pyDE1.event_manager import SubscribedEvent
 from pyDE1.de1.events import ShotSampleUpdate, \
     ShotSampleWithVolumesUpdate
+from pyDE1.scanner import DiscoveredDevices, BleakScannerWrapped
 from pyDE1.singleton import Singleton
 
 from pyDE1.utils import task_name_exists, cancel_tasks_by_name
 
 from pyDE1.dispatcher.resource import ConnectivityEnum, DE1ModeEnum
 
+from pyDE1.config.bluetooth import CONNECT_TIMEOUT
+
+from pyDE1.scanner import find_first_matching
+
 
 # If True, randomly skips upload packets
 _TEST_BLE_LOSS_DURING_FW_UPLOAD = False
 if _TEST_BLE_LOSS_DURING_FW_UPLOAD:
     import random
 
-logger = logging.getLogger('de1')
-
-# TODO: Initialization should be able to be done by address or BLEDevice
-#       NB: https://github.com/hbldh/bleak/issues/361 on Linux and scan
-
-# TODO: Scanner that can return first quickly or all within a time period
-
-# NB: It doesn't look like CoreBluetooth reveals an identifier that is stable
-#     across boots.  It uses a UUID rather than the MAC address of the device.
+logger = logging.getLogger('DE1')
 
 class DE1 (Singleton):
 
@@ -88,8 +84,10 @@ class DE1 (Singleton):
     # def __init__(self):
     #     pass
 
+    MAX_WAIT_FOR_READY_EVENTS = 3.0  # seconds, 1.5-2.5 seconds seems typical
+
     def _singleton_init(self):
-        self._address = None
+        self._address_or_bledevice: Optional[Union[str, BLEDevice]] = None
         self._name = None
         self._bleak_client: Optional[BleakClientWrapped] = None
 
@@ -100,14 +98,11 @@ class DE1 (Singleton):
         # TODO: Should the handlers be able to be changed on the fly?
         self._handlers = pyDE1.de1.handlers.default_handler_map(self)
 
-        # The _cuuid_dict and _mmr_dict dictionaries contain the last update
         # TODO: Where is "last heard from"?
         self._cuuid_dict: Dict[CUUID, NotificationState] = dict()
         self._mmr_dict: Dict[Union[MMR0x80LowAddr, int], MMR0x80Data] = dict()
-        for cuuid in CUUID:
-            self._cuuid_dict[cuuid] = NotificationState(cuuid)
-        for mmr in MMR0x80LowAddr:
-            self._mmr_dict[mmr] = MMR0x80Data(mmr)
+
+        self._latest_profile: Optional[Profile] = None
 
         self._event_connectivity = SubscribedEvent(self)
         self._event_state_update = SubscribedEvent(self)
@@ -115,11 +110,7 @@ class DE1 (Singleton):
         self._event_water_levels = SubscribedEvent(self)
         self._event_shot_sample_with_volumes_update = SubscribedEvent(self)
 
-        # These have to be valid, as may be tested before first update
-        self._cuuid_dict[CUUID.StateInfo]._last_value = StateInfo(
-            State=API_MachineStates.NoRequest,
-            SubState=API_Substates.NoState
-        )
+        self._ready = asyncio.Event()
 
         # Used to restrict multiple access to writing the active profile
         self._profile_lock = asyncio.Lock()
@@ -144,13 +135,130 @@ class DE1 (Singleton):
         asyncio.create_task(self._event_shot_sample.subscribe(
             self._create_self_callback_ssu()))
 
-    # TODO: Rework initialization
+        self.prepare_for_connection(wipe_address=True)
+
+    #
+    # High-level initialization and re-initialization
+    #
+
+    def prepare_for_connection(self, wipe_address=True):
+        """
+        Basically wipe all cached state
+        """
+        if self.is_connected:
+            raise DE1IsConnectedError(
+                "Can't prepare_for_connection() while connected.")
+
+        logger.info(
+            f"prepare_for_connection(wipe_address={wipe_address})")
+
+        loop = asyncio.get_running_loop()
+        if loop is not None and loop.is_running():
+            asyncio.create_task(self._notify_not_ready())
+        else:
+            logger.debug(f"No running loop to _notify_not_ready(): {loop}")
+
+        if wipe_address:
+            self._address_or_bledevice = None
+            self._bleak_client = None  # Constructor requires an address
+
+        self._cuuid_dict: Dict[CUUID, NotificationState] = dict()
+        self._mmr_dict: Dict[Union[MMR0x80LowAddr, int], MMR0x80Data] = dict()
+        for cuuid in CUUID:
+            self._cuuid_dict[cuuid] = NotificationState(cuuid)
+        for mmr in MMR0x80LowAddr:
+            self._mmr_dict[mmr] = MMR0x80Data(mmr)
+
+        self._cuuid_dict[CUUID.StateInfo]._last_value = StateInfo(
+            State=API_MachineStates.NoRequest,
+            SubState=API_Substates.NoState
+        )
+
+        self._tracking_volume_dispensed = False
+        self._volume_dispensed_total = 0
+        self._volume_dispensed_preinfuse = 0
+        self._volume_dispensed_pour = 0
+        self._volume_dispensed_by_frame = []
+        # TODO: Convince Ray to return substate and state
+        #       in ShotSample so this isn't needed for volume tracking
+        self._number_of_preinfuse_frames: int = 0
+
+        self._latest_profile = None
+
+        self._last_stop_requested = 0
+
+        # Internal flag
+        self._recorder_active = False
+
+    async def initialize_after_connection(self):
+
+        logger.info("initialize_after_connection()")
+
+        await asyncio.gather(
+            self.start_standard_read_write_notifiers(),
+            self.start_standard_periodic_notifiers(),
+        )
+
+        t0 = time.time()
+        ((event_list, addr_low_list), ignore, ignore, ignore) \
+            = await asyncio.gather(
+            self.read_standard_mmr_registers(),
+            self.read_cuuid(CUUID.StateInfo),
+            self.read_cuuid(CUUID.Versions),
+            self.read_cuuid(CUUID.ShotSettings)
+        )
+
+        event_list.append(self._cuuid_dict[CUUID.StateInfo].ready_event)
+
+        # TODO: How to detect all the info has come back
+        #       and what to do if it doesn't?
+
+        gather_list = [event.wait() for event in event_list]
+
+        # TODO: Timeout!!
+        logger.info(f"Waiting for {len(event_list)} responses")
+        try:
+            results = await asyncio.wait_for(asyncio.gather(*gather_list),
+                             self.MAX_WAIT_FOR_READY_EVENTS)
+            t1 = time.time()
+            logger.info(
+                f"{len(event_list)} responses received in "
+                f"{t1 - t0:.3f} seconds")
+        except asyncio.TimeoutError:
+            logger.warning("Timeout waiting for responses.")
+            idx = 0
+            for event in event_list:
+                event: asyncio.Event
+                if not event.is_set():
+                    if idx < len(event_list) - 1:
+                        failed = MMR0x80LowAddr(addr_low_list[idx])
+                        logger.warning(
+                            f"No response from #{idx + 1} "
+                            f"of {len(event_list)}, " \
+                            + str(failed))
+                        await self.read_one_mmr0x80(failed)
+                    else:
+                        logger.warning(
+                            "No response from CUUID.StateInfo"
+                        )
+                        await self.read_cuuid(CUUID.StateInfo)
+                idx += 1
+            logger.error("Stupidly continuing anyway after re-requesting")
+
+        # NB: This gets sent if all there or not
+        await self._notify_ready()
+
+        return
+
     async def _notify_ready(self):
+        self._ready.set()
         await self._event_connectivity.publish(
             ConnectivityChange(arrival_time=time.time(),
                                state=ConnectivityState.READY))
+        logger.info("Ready")
 
     async def _notify_not_ready(self):
+        self._ready.clear()
         await self._event_connectivity.publish(
             ConnectivityChange(arrival_time=time.time(),
                                state=ConnectivityState.NOT_READY))
@@ -165,27 +273,86 @@ class DE1 (Singleton):
 
     @property
     def address(self):
-        addr = self._address
+        addr = self._address_or_bledevice
         if isinstance(addr, BLEDevice):
             addr = addr.address
         return addr
 
-    @address.setter
-    def address(self, address: Union[BLEDevice, str]):
-        if self.address is not None:
-            raise DE1APIValueError(
-                "Changing the DE1 address is not yet supported")
-        # If using BLEDevice directly
-        # AttributeError: 'BleakClientBlueZDBus' object has no attribute 'lower'
-        if isinstance(address, BleakClientBlueZDBus):
-            self._address = address.address
+    async def set_address(self, address: Optional[Union[BLEDevice, str]]):
+        if address is None:
+            await self.disconnect()
+            self.prepare_for_connection(wipe_address=True)
+        elif address == self.address:
+            pass
         else:
-            self._address = address
+            await self.disconnect()
+            self.prepare_for_connection(wipe_address=True)
+            self._address_or_bledevice = address
+            if isinstance(address, BLEDevice):
+                self._name = address.name
+            self._bleak_client = BleakClientWrapped(self._address_or_bledevice)
+
+    #
+    # Self-contained calls for API
+    #
+
+    async def first_if_found(self, doit: bool):
+        if self.is_connected:
+            logger.warning(
+                "first_if_found requested, but already connected. "
+                "No action taken.")
+        elif not doit:
+            logger.warning(
+                "first_if_found requested, but not True. No action taken.")
+        else:
+            device = await find_first_matching(('DE1',))
+            if device:
+                await self.change_de1_to_id(device.address)
+        return self.address
+
+    async def change_de1_to_id(self, ble_device_id: Optional[str]):
+        """
+        For now, this won't return until connected or fails to connect
+        As a result, will trigger the timeout on API calls
+        """
+        logger.info(f"Address change requested for DE1 from {self.address} "
+                    f"to {ble_device_id}")
+
+        de1 = DE1()
+
+        # TODO: Need to make distasteful assumption that the id is the address
+        try:
+            if self.address == ble_device_id:
+                logger.info(f"Already using {ble_device_id}. No action taken")
+                return
+        except AttributeError:
+            pass
+
+        if ble_device_id is None:
+            # Straightforward request to disconnect and not replace
+            await self.set_address(None)
+
+        else:
+            ble_device = DiscoveredDevices().ble_device_from_id(ble_device_id)
+            if ble_device is None:
+                logger.warning(f"No record of {ble_device_id}, initiating scan")
+                # TODO: find_device_by_filter doesn't add to DiscoveredDevices
+                ble_device = await BleakScannerWrapped.find_device_by_address(
+                    ble_device_id, timeout=CONNECT_TIMEOUT)
+            if ble_device is None:
+                raise DE1NoAddressError(
+                    f"Unable to find device with id: '{ble_device_id}'")
+            await self.set_address(ble_device)
+            await self.connect()
 
     @property
     def name(self):
         return self._name
 
+    @property
+    def latest_profile(self):
+        return deepcopy(self._latest_profile)
+
     @property
     def event_state_update(self):
         return self._event_state_update
@@ -202,34 +369,50 @@ class DE1 (Singleton):
     def event_shot_sample_with_volumes_update(self):
         return self._event_shot_sample_with_volumes_update
 
+    @property
+    def allow_flow_start_commands(self):
+        ghc_info = self._mmr_dict[MMR0x80LowAddr.GHC_INFO].data_decoded
+        if ghc_info is None or ghc_info & MMRGHCInfoBitMask.GHC_ACTIVE:
+            return False
+        else:
+            return True
+
     # NB: Linux apparently "needs" a scan when connecting by address
     #     It may be the case that this disconnects other devices
     #     See: https://github.com/hbldh/bleak/issues/361
 
-    async def connect(self, timeout=5.0):
+    async def connect(self, timeout: Optional[float] = None):
+
+        if timeout is None:
+            timeout = CONNECT_TIMEOUT
+
         logger.info(f"Connecting to DE1 at {self.address}")
-        if self._bleak_client is None:
-            if self.address is None:
-                raise DE1NoAddressError
-            self._bleak_client = BleakClientWrapped(self.address)
+
+        assert self._bleak_client is not None
         self._bleak_client.set_disconnected_callback(
             self._create_disconnect_callback()
         )
+
         await asyncio.gather(self._event_connectivity.publish(
             ConnectivityChange(arrival_time=time.time(),
                                state=ConnectivityState.CONNECTING)),
             self._bleak_client.connect(timeout=timeout),
         )
+
         if self.is_connected:
-            self._address = self._bleak_client.address
             if self.name is None:
                 self._name = self._bleak_client.name
             logger.info(f"Connected to DE1 at {self.address}")
             await self._event_connectivity.publish(
                 ConnectivityChange(arrival_time=time.time(),
                                    state=ConnectivityState.CONNECTED))
+            # This can take some time, potentially delaying scale connection
+            # At least BlueZ doesn't like concurrent connection requests
+            asyncio.create_task(self.initialize_after_connection())
+
         else:
             logger.error(f"Connection failed to DE1 at {self.address}")
+            await self._notify_not_ready()
             await self._event_connectivity.publish(
                 ConnectivityChange(arrival_time=time.time(),
                                    state=ConnectivityState.DISCONNECTED))
@@ -237,8 +420,10 @@ class DE1 (Singleton):
     async def disconnect(self):
         logger.info(f"Disconnecting from DE1")
         if self._bleak_client is None:
-            logger.info(f"Disconnecting from DE1; no client")
+            logger.info(f"Disconnecting from DE1, no client")
+            return
         await asyncio.gather(
+            self._notify_not_ready(),
             self._bleak_client.disconnect(),
             self._event_connectivity.publish(
                 ConnectivityChange(arrival_time=time.time(),
@@ -255,7 +440,6 @@ class DE1 (Singleton):
                 ConnectivityChange(arrival_time=time.time(),
                                    state=ConnectivityState.DISCONNECTED))
 
-
     # TODO: Decide how to handle  self._disconnected_callback
     #   disconnected_callback (callable): Callback that will be scheduled in the
     #   event loop when the client is disconnected. The callable must take one
@@ -275,9 +459,16 @@ class DE1 (Singleton):
                 f"Disconnected from DE1 at {client.address}, "
                 "willful_disconnect: "
                 f"{client.willful_disconnect}")
-            asyncio.create_task(de1._event_connectivity.publish(
-                ConnectivityChange(arrival_time=time.time(),
-                                   state=ConnectivityState.DISCONNECTED)))
+
+            # asyncio.gather is a Future, not a Task
+            asyncio.ensure_future(asyncio.gather(
+                self._notify_not_ready(),
+                de1._event_connectivity.publish(
+                    ConnectivityChange(arrival_time=time.time(),
+                                       state=ConnectivityState.DISCONNECTED)),
+            ))
+            # TODO: if not shutdown underway:
+            de1.prepare_for_connection(wipe_address=client.willful_disconnect)
 
         return disconnect_callback
 
@@ -480,7 +671,8 @@ class DE1 (Singleton):
 # MMR-based properties
 #
 
-    async def read_standard_mmr_registers(self):
+    async def read_standard_mmr_registers(self) -> (List[asyncio.Event],
+                                                    List[MMR0x80LowAddr]):
         """
         Request a read of the readable MMR registers, in bulk
         :return:
@@ -508,9 +700,21 @@ class DE1 (Singleton):
         #     self.read_mmr(words_block_3 - 1, 0x80, start_block_3),
         # )
 
-        await self.read_mmr(words_block_1, 0x80, start_block_1)
-        await self.read_mmr(words_block_2, 0x80, start_block_2)
-        await self.read_mmr(words_block_3, 0x80, start_block_3)
+        event_list = await self.read_mmr(
+            words_block_1, 0x80, start_block_1)
+        event_list.extend(await self.read_mmr(
+            words_block_2, 0x80, start_block_2))
+        event_list.extend(await self.read_mmr(
+            words_block_3, 0x80, start_block_3))
+
+        addr_low_list = []
+        addr_low_list.extend(range(start_block_1, end_block_1 + 4, 4))
+        addr_low_list.extend(range(start_block_2, end_block_2 + 4, 4))
+        addr_low_list.extend(range(start_block_3, end_block_3 + 4, 4))
+
+        return event_list, addr_low_list
+
+
 
     #
     # Upload a shot profile
@@ -521,17 +725,11 @@ class DE1 (Singleton):
     async def upload_json_v2_profile(self, profile: Union[bytes,
                                                           bytearray,
                                                           str]):
-        if isinstance(profile, (bytes, bytearray)):
-            profile = profile.decode('utf-8')
-        elif isinstance(profile, str):
-            pass
-        else:
-            raise DE1APITypeError(
-                "Profile should be bytes, bytearray, or str, "
-                f"not {type(profile)}")
-        pj = json.loads(profile)
-        pbf = ProfileByFrames().from_json(pj)
+        pbf = ProfileByFrames().from_json(profile)
         await self.upload_profile(pbf)
+        async with aiosqlite.connect('/var/lib/pyDE1/pyDE1.sqlite3') as db:
+            await db_insert.profile(pbf, db, time.time())
+            logger.info("Returned from db insert")
 
     # "Internal" version
 
@@ -556,14 +754,17 @@ class DE1 (Singleton):
                 raise DE1OperationInProgressError(
                     'There is already a profile upload in progress')
         profile_upload_stopped = asyncio.Event()
-        await asyncio.create_task(self._upload_profile(
+        upload_task = asyncio.create_task(self._upload_profile(
             profile=profile,
             override_stop_limits=osl,
             override_tank_temperature=ott,
             profile_upload_stopped=profile_upload_stopped)
         )
         await profile_upload_stopped.wait()
-        # TODO: Not clear here if interrupted or successful
+        if (e := upload_task.exception()) is not None:
+            logger.info(f"Upload task exception: {upload_task.exception()}")
+            # TODO: This will raise on cancelled, is that "right"?
+            raise e
 
 
     @staticmethod
@@ -589,12 +790,23 @@ class DE1 (Singleton):
                         done = await self.start_notifying(cuuid)
                         # await done.wait()
 
+                self._latest_profile = None
+                bytes_for_fingerprint = bytearray()
+
                 await self.write_packed_attr(profile.header_write())
+                bytes_for_fingerprint += profile.header_write().as_wire_bytes()
                 for frame in profile.shot_frame_writes():
                     await self.write_packed_attr(frame)
+                    bytes_for_fingerprint += frame.as_wire_bytes()
                 for frame in profile.ext_shot_frame_writes():
                     await self.write_packed_attr(frame)
+                    bytes_for_fingerprint += frame.as_wire_bytes()
                 await self.write_packed_attr(profile.shot_tail_write())
+                bytes_for_fingerprint \
+                    += profile.shot_tail_write().as_wire_bytes()
+
+                profile._fingerprint = hashlib.sha1(
+                    bytes_for_fingerprint).hexdigest()
 
                 if profile.number_of_preinfuse_frames is not None:
                     self._number_of_preinfuse_frames = \
@@ -625,6 +837,7 @@ class DE1 (Singleton):
                             weight=target
                         )
 
+                self._latest_profile = profile
 
         except asyncio.CancelledError:
             pass
@@ -903,7 +1116,11 @@ class DE1 (Singleton):
                     else:
                         de1._volume_dispensed_preinfuse += v_inc
                     de1._volume_dispensed_total += v_inc
-                    de1._volume_dispensed_by_frame[ssu.frame_number] += v_inc
+                    if de1.current_state is API_MachineStates.Espresso:
+                        to_frame = ssu.frame_number
+                    else:
+                        to_frame = 0
+                    de1._volume_dispensed_by_frame[to_frame] += v_inc
 
             last_sample_time = ssu.sample_time
 
@@ -980,40 +1197,68 @@ class DE1 (Singleton):
     # For API
     @property
     def connectivity(self):
+        retval = ConnectivityEnum.NOT_CONNECTED
         if self.is_connected:
-            retval = ConnectivityEnum.CONNECTED
-        else:
-            # intentionally vague, as PUT "connecting" isn't needed
-            retval = ConnectivityEnum.NOT_CONNECTED
+            if self._ready.is_set():
+                retval = ConnectivityEnum.READY
+            else:
+                retval = ConnectivityEnum.CONNECTED
         return retval
 
     async def connectivity_setter(self, value):
         assert isinstance(value, ConnectivityEnum), \
-            f"mode of {value} not a ConnectivityEnum " \
-            "in DE1.connectivity_setter()"
-        if value is ConnectivityEnum.CONNECTED \
-                and not self.is_connected:
+            f"mode of {value} not a ConnectivityEnum "
+        if value is ConnectivityEnum.CONNECTED and not self.is_connected:
             await self.connect()
-        elif value is ConnectivityEnum.NOT_CONNECTED \
-                and self.is_connected:
+        elif value is ConnectivityEnum.NOT_CONNECTED and self.is_connected:
             await self.disconnect()
+        else:
+            raise DE1APIValueError(
+                "Only CONNECTED and NOT_CONNECTED can be set, "
+                f"not {value}")
 
     @property
     def auto_off_time(self):
         return None
 
+    # Non-GHC only -- no checking for GHC presence at this level
+
+    async def _flow_start_espresso(self):
+        await self.write_packed_attr(RequestedState(
+            State=API_MachineStates.Espresso))
+
+    async def _flow_start_hot_water_rinse(self):
+        await self.write_packed_attr(RequestedState(
+            State=API_MachineStates.HotWaterRinse))
+
+    async def _flow_start_steam(self):
+        await self.write_packed_attr(RequestedState(
+            State=API_MachineStates.Steam))
+
+    async def _flow_start_hot_water(self):
+        await self.write_packed_attr(RequestedState(
+            State=API_MachineStates.HotWater))
+
+    # General mode activation -- check is here for GHC or not
+
     async def mode_setter(self, mode: DE1ModeEnum):
         assert isinstance(mode, DE1ModeEnum), \
             f"mode of {mode} not a DE1ModeEnum in DE1.mode_setter()"
+
+        # Ensure GHC data has been read
+        if self._mmr_dict[MMR0x80LowAddr.GHC_INFO].data_decoded is None:
+            logger.info("GHC_INFO not present, reading now.")
+            await self.read_one_mmr0x80_and_wait(MMR0x80LowAddr.GHC_INFO)
+
         cs = self.current_state
         if cs is API_MachineStates.NoRequest:
             logger.warning(f"Refreshing current state as is NoRequest")
             await self.read_cuuid(CUUID.StateInfo)
             cs = self.current_state
+        css = self.current_substate
         logger.debug(f"Request to change mode to {mode} "
                      f"while in {API_MachineStates(cs).name}")
 
-
         if mode is DE1ModeEnum.SLEEP:
             if cs is API_MachineStates.Idle:
                 logger.debug("API triggered sleep()")
@@ -1022,7 +1267,7 @@ class DE1 (Singleton):
                                       API_MachineStates.GoingToSleep):
                 pass
             else:
-                raise DE1APIUnsupportedStateTransitionError (mode, cs)
+                raise DE1APIUnsupportedStateTransitionError(mode, cs, css)
 
         elif mode is DE1ModeEnum.WAKE:
             logger.debug(f"In WAKE, cs: {API_MachineStates(cs).name}")
@@ -1048,28 +1293,51 @@ class DE1 (Singleton):
                         API_MachineStates.SkipToNext,
                         API_MachineStates.Refill,
                         API_MachineStates.InBootLoader):
-                raise DE1APIUnsupportedStateTransitionError(mode, cs)
+                raise DE1APIUnsupportedStateTransitionError(mode, cs, css)
             else:
                 logger.debug("API triggered idle()")
                 await self.idle()
 
         elif mode is DE1ModeEnum.SKIP_TO_NEXT:
 
-            if cs in (API_MachineStates.Espresso) \
+            if cs is API_MachineStates.Espresso \
                 and self.current_substate in (API_Substates.PreInfuse,
                                               API_Substates.Pour):
                 logger.debug("API triggered skip_to_next()")
                 await self.skip_to_next()
             else:
-                raise DE1APIUnsupportedStateTransitionError(
-                    mode, f"{cs}, {self.current_substate}")
+                raise DE1APIUnsupportedStateTransitionError(mode, cs, css)
 
+        elif mode in (DE1ModeEnum.ESPRESSO,
+                      DE1ModeEnum.HOT_WATER_RINSE,
+                      DE1ModeEnum.STEAM,
+                      DE1ModeEnum.HOT_WATER,
+                      ):
 
+            if not self.allow_flow_start_commands:
+                raise DE1APIUnsupportedFeatureError(
+                    f"DE1 does not permit {mode} unless GHC is not installed."
+                )
 
+            if cs is not API_MachineStates.Idle:
+                raise DE1APIUnsupportedStateTransitionError(mode, cs, css)
 
+            if mode is DE1ModeEnum.ESPRESSO:
+                logger.debug("API triggered _flow_start_espresso()")
+                await self._flow_start_espresso()
 
+            elif mode is DE1ModeEnum.HOT_WATER_RINSE:
+                logger.debug("API triggered _flow_start_hot_water_rinse()")
+                await self._flow_start_hot_water_rinse()
 
+            elif mode is DE1ModeEnum.STEAM:
+                logger.debug("API triggered _flow_start_steam()")
+                await self._flow_start_steam()
 
+            elif mode is DE1ModeEnum.HOT_WATER:
+                logger.debug("API triggered _flow_start_hot_water()")
+                await self._flow_start_hot_water()
+
+        else:
+            raise DE1APIUnsupportedStateTransitionError(mode, cs, css)
 
-# from pyDE1.dispatcher.mapping import MAPPING
-# from pyDE1.dispatcher.implementation import get_target_sets
