@@ -13,8 +13,9 @@ Provide "mock" classes for use in inbound API process
 
 NB: Must be imported AFTER any imports of DE1, Scale, FlowSequencer, ...
 """
+import importlib.util
 import logging
-import multiprocessing, multiprocessing.connection
+import multiprocessing
 import sys
 from typing import Optional, Union, NamedTuple
 
@@ -23,46 +24,52 @@ from pyDE1.dispatcher.resource import Resource, RESOURCE_VERSION, \
 
 # TODO: Work through main and remote thread imports
 
-logger = logging.getLogger(multiprocessing.current_process().name)
+logger = logging.getLogger('Mapping')
 
 pname = f"into {multiprocessing.current_process().name} process"
 
 if 'pyDE1.de1' in sys.modules:
+    logger.info("Importing DE1")
     from pyDE1.de1 import DE1
 else:
-    print("Importing stub for DE1", pname)
-    logger.info("Importing stub for DE1")
+    logger.info("Using stub for DE1")
     from pyDE1.dispatcher.stubs import DE1
 
 if 'pyDE1.scale' in sys.modules:
+    logger.info("Importing Scale")
     from pyDE1.scale import Scale
 else:
-    print("Importing stub for Scale", pname)
-    logger.info("Importing stub for Scale")
+    logger.info("Using stub for Scale")
     from pyDE1.dispatcher.stubs import Scale
 
+if 'pyDE1.scale.processor' in sys.modules:
+    logger.info("Importing ScaleProcessor")
+    from pyDE1.scale.processor import ScaleProcessor
+else:
+    logger.info("Using stub for ScaleProcessor")
+    from pyDE1.dispatcher.stubs import ScaleProcessor
+
 if 'pyDE1.flow_sequencer' in sys.modules:
+    logger.info("Importing FlowSequencer")
     from pyDE1.flow_sequencer import FlowSequencer
 else:
-    print("Importing stub for FlowSequencer", pname)
-    logger.info("Importing stub for FlowSequencer")
+    logger.info("Using stub for FlowSequencer")
     from pyDE1.dispatcher.stubs import FlowSequencer
 
-# cpn = multiprocessing.current_process().name
-# for k in sys.modules.keys():
-#     if (k.startswith('pyDE1')
-#             or k.startswith('bleak')
-#             or k.startswith('asyncio-mqtt')):
-#         print(
-#             f"{cpn}: EARLY: {k}"
-#         )
+if 'pyDE1.scanner' in sys.modules:
+    logger.info("Importing DiscoveredDevices")
+    from pyDE1.scanner import DiscoveredDevices
+else:
+    logger.info("Using stub for DiscoveredDevices")
+    from pyDE1.dispatcher.stubs import DiscoveredDevices
+
 
 from pyDE1.de1.c_api import PackedAttr, MMR0x80LowAddr, get_cuuid, \
     ShotSettings, SetTime, Versions, WaterLevels
 
-from pyDE1.de1.exceptions import DE1APIValueError
+from pyDE1.exceptions import DE1APIValueError
 
-MAPPING_VERSION = "2.1.0"
+MAPPING_VERSION = "3.1.0"
 
 
 class IsAt (NamedTuple):
@@ -167,7 +174,10 @@ import importlib.metadata  # Used for module-version lookup only
 def module_versions():
     retval = {}
     for module in MODULES_FOR_VERSIONS:
-        retval[module] = importlib.metadata.version(module)
+        try:
+            retval[module] = importlib.metadata.version(module)
+        except importlib.metadata.PackageNotFoundError:
+            retval[module] = None
     return retval
 
 
@@ -194,14 +204,33 @@ MAPPING[Resource.DE1_PROFILE] = IsAt(target=DE1, attr_path=None,
                                      setter_path='upload_json_v2_profile',
                                      v_type=Union[bytes, bytearray])
 
+# TODO: How to reference a module-level subroutine? Use module as target?
+# TODO: How to get the scan ID back to the caller
+
+MAPPING[Resource.SCAN] = {
+    'begin': IsAt(target=None, attr_path=None,
+                  setter_path='scan_from_api', v_type=bool)
+}
+
+# TODO: This is an async getter because of the lock
+MAPPING[Resource.SCAN_DEVICES] = {
+    'devices': IsAt(target=DiscoveredDevices, attr_path='devices_for_json',
+                    read_only=True,
+                    v_type=dict)
+}
+
 # Work from leaves back, so can be "included" by reference
 
-# For now, these are not writable
+# For now, name is not writable
 MAPPING[Resource.DE1_ID] = {
     'name': IsAt(target=DE1, attr_path='name', v_type=str,
                  read_only=True),
-    'id': IsAt(target=DE1, attr_path='address', v_type=str,
-               read_only=True),
+    'id': IsAt(target=DE1, attr_path='address', v_type=Optional[str],
+               setter_path='change_de1_to_id'),
+    # first_if_found, if true, will replace only if one is found
+    # It is an error to be true if 'id' is present at this time
+    'first_if_found': IsAt(target=DE1, attr_path=None,
+                        setter_path='first_if_found', v_type=bool),
 }
 
 # NB: A single-entry tuple needs to end with a comma
@@ -406,10 +435,15 @@ MAPPING[Resource.DE1_CALIBRATION_LINE_FREQUENCY] = {
 MAPPING[Resource.SCALE_ID] = {
     'name': IsAt(target=Scale, attr_path='name', v_type=str,
                  read_only=True),
-    'id': IsAt(target=Scale, attr_path='address', v_type=str,
-               read_only=True),
+    'id': IsAt(target=ScaleProcessor, attr_path='scale_address',
+               v_type=Optional[str],
+               setter_path='change_scale_to_id'),
     'type': IsAt(target=Scale, attr_path='type', v_type=str,
                  read_only=True),
+    # first_if_found, if true, will replace only if one is found
+    # It is an error to be true if 'id' is present at this time
+    'first_if_found': IsAt(target=ScaleProcessor, attr_path=None,
+                           setter_path='first_if_found', v_type=bool),
 }
 
 MAPPING[Resource.SCALE_CONNECTIVITY] = {
