@@ -12,6 +12,11 @@ class DE1:
 class Scale:
     pass
 
+class ScaleProcessor:
+    pass
+
 class FlowSequencer:
     pass
 
+class DiscoveredDevices:
+    pass
\ No newline at end of file
