@@ -11,18 +11,20 @@ FlowSequencer
 Responsible for coordinating the actions around any of the flow modes
 """
 import enum
+import multiprocessing
 import time
 
 import asyncio
 import logging
+import warnings
 
 from typing import Optional, Callable, Coroutine, List
 
 from pyDE1.de1 import DE1
 from pyDE1.de1.events import ShotSampleUpdate, StateUpdate, \
     ShotSampleWithVolumesUpdate
-from pyDE1.de1.c_api import API_MachineStates, API_Substates, CUUID
-from pyDE1.de1.exceptions import DE1APIError, DE1APITypeError, DE1APIValueError, \
+from pyDE1.de1.c_api import API_MachineStates, API_Substates
+from pyDE1.exceptions import DE1APIValueError, \
     DE1APIAttributeError
 
 from pyDE1.i_target_setter import I_TargetSetter
@@ -30,8 +32,8 @@ from pyDE1.i_target_setter import I_TargetSetter
 from pyDE1.scale.processor import ScaleProcessor
 from pyDE1.scale.events import WeightAndFlowUpdate
 
-from pyDE1.event_manager import EventWithNotify, EventNotificationName, \
-    EventPayload, send_to_outbound_queue
+from pyDE1.event_manager import SequencerGate, SequencerGateName, \
+    SequencerGateNotification, EventPayload, send_to_outbound_pipes
 from pyDE1.singleton import Singleton
 
 from pyDE1.utils import cancel_tasks_by_name
@@ -40,6 +42,9 @@ from pyDE1.flow_sequencer.mode_control import ByModeControl, \
     StopAtTime, StopAtWeight, StopAtVolume,  \
     EspressoControl, SteamControl, HotWaterControl, HotWaterRinseControl
 
+# import pyDE1.database.write_notifications import create_history_record
+import pyDE1.database.write_notifications
+
 logger = logging.getLogger('FlowSequencer')
 
 
@@ -110,37 +115,41 @@ class FlowSequencer (Singleton, I_TargetSetter):
     # def __init__(self):
     #     pass
 
-    def _singleton_init(self):
-        self._de1: Optional[DE1] = None
-        self._scale_processor: Optional[ScaleProcessor] = None
+    database_queue: Optional[multiprocessing.Queue] = None
 
-        self._event_de1_present = asyncio.Event()
-        self._event_scale_processor_present = asyncio.Event()
+    def _singleton_init(self):
+
+        # Singletons now
+        self._de1 = DE1()
+        self._scale_processor = ScaleProcessor()
+
+        # self._event_de1_present = asyncio.Event()
+        # self._event_scale_processor_present = asyncio.Event()
 
         self._active_state: Optional[API_MachineStates] = None
         self._sequence_start_time: float = 0
         # For later cancellation. Should this be static or dynamic?
         self._sequence_task_name: str = "SequencerSubtask"
 
-        self._gate_sequence_start = EventWithNotify(self,
-            EventNotificationName.GATE_SEQUENCE_START)
+        self._gate_sequence_start = SequencerGate(self,
+            SequencerGateName.GATE_SEQUENCE_START)
         # These are not necessarily in order
-        self._gate_flow_begin = EventWithNotify(self,
-            EventNotificationName.GATE_FLOW_BEGIN)
-        self._gate_expect_drops = EventWithNotify(self,
-            EventNotificationName.GATE_EXPECT_DROPS)
+        self._gate_flow_begin = SequencerGate(self,
+            SequencerGateName.GATE_FLOW_BEGIN)
+        self._gate_expect_drops = SequencerGate(self,
+            SequencerGateName.GATE_EXPECT_DROPS)
         # self._gate_first_drops = asyncio.Event()
-        self._gate_exit_preinfuse = EventWithNotify(self,
-            EventNotificationName.GATE_EXIT_PREINFUSE)
-        self._gate_flow_end = EventWithNotify(self,
-            EventNotificationName.GATE_FLOW_END)
-        self._gate_flow_state_exit = EventWithNotify(self,
-            EventNotificationName.GATE_FLOW_STATE_EXIT)
-        self._gate_last_drops = EventWithNotify(self,
-            EventNotificationName.GATE_LAST_DROPS)
+        self._gate_exit_preinfuse = SequencerGate(self,
+            SequencerGateName.GATE_EXIT_PREINFUSE)
+        self._gate_flow_end = SequencerGate(self,
+            SequencerGateName.GATE_FLOW_END)
+        self._gate_flow_state_exit = SequencerGate(self,
+            SequencerGateName.GATE_FLOW_STATE_EXIT)
+        self._gate_last_drops = SequencerGate(self,
+            SequencerGateName.GATE_LAST_DROPS)
         # Though always "done" here (and should always be triggered)
-        self._gate_sequence_complete = EventWithNotify(self,
-            EventNotificationName.GATE_SEQUENCE_COMPLETE)
+        self._gate_sequence_complete = SequencerGate(self,
+            SequencerGateName.GATE_SEQUENCE_COMPLETE)
 
         self._all_gates = [
             self._gate_sequence_start,
@@ -193,15 +202,17 @@ class FlowSequencer (Singleton, I_TargetSetter):
         self._stop_at_volume_active = False
         self._stop_at_time_active = False
 
+        asyncio.create_task(self.set_up_subscribers())
+
     @property
     def de1(self):
         return self._de1
 
-    async def set_de1(self, value: DE1):
-        # Needs to unsubscribe the old
-        self._de1 = value
+    async def set_up_subscribers(self):
+        # DE1 only knows that its _flow_sequencer is an I_TargetSetter
+        # because of circular imports, so it can't reference FlowSequencer
         self._de1._flow_sequencer = self
-        self._event_de1_present.set()
+
         await asyncio.gather(
             self.de1.event_state_update.subscribe(
                 self._create_state_update_callback()),
@@ -212,37 +223,16 @@ class FlowSequencer (Singleton, I_TargetSetter):
             self.de1.event_shot_sample_with_volumes_update.subscribe(
                 self._create_stop_at_volume_subscriber()),
 
-            # This will fail if the scale_processor hasn't been connected yet
-            # self.scale_processor.event_weight_and_flow_update.subscribe(
-            #     _create_stop_at_weight_subscriber(self.de1)),
+            self.scale_processor.event_weight_and_flow_update.subscribe(
+                self._create_stop_at_weight_subscriber()),
         )
-        asyncio.create_task(self._set_stop_at_weight_subscriber_when_possible())
+        logger.info("FlowSequencer subscriptions done")
         return self
 
-    #
-    # TODO: THIS NEEDS A LOT OF HELP FOR CHANGING EITHER
-    #
-    async def _set_stop_at_weight_subscriber_when_possible(self):
-        await asyncio.gather(
-            self._event_scale_processor_present.wait(),
-            self._event_de1_present.wait(),
-        )
-        await self.scale_processor.event_weight_and_flow_update.subscribe(
-            self._create_stop_at_weight_subscriber()
-        )
-
     @property
     def scale_processor(self):
         return self._scale_processor
 
-    async def set_scale_processor(self, value: ScaleProcessor):
-        # Needs to unsubscribe the old
-        self._scale_processor = value
-        self._event_scale_processor_present.set()
-        await self._scale_processor.event_weight_and_flow_update.subscribe(
-                self._create_weight_and_flow_update_callback())
-        return self
-
     @property
     def active_state(self):
         return self._active_state
@@ -349,6 +339,7 @@ class FlowSequencer (Singleton, I_TargetSetter):
 
         return flow_sequencer_ssu_cb
 
+    # Should this be async def now?
     def _start_sequence(self, state: API_MachineStates):
         """
         Kick off parallel tasks to manage functions during the shot
@@ -362,6 +353,14 @@ class FlowSequencer (Singleton, I_TargetSetter):
             self._abort_sequence()
 
         self._active_state = state
+
+        sequence_id = SequencerGateNotification.new_sequence()
+        logger.info(f"Starting sequence_id {sequence_id}")
+
+        # Can't really start the recorder here as it needs
+        # to be able to create the sequence (history) record
+        # which can take up to 250 ms
+
         self._close_all_gates()
 
         # create a bunch of tasks
@@ -405,7 +404,7 @@ class FlowSequencer (Singleton, I_TargetSetter):
         Maybe a funky way to do it, but this captures "done-ness"
         in one place.
         """
-        await self._gate_flow_state_exit.wait()
+        await self._gate_flow_end.wait()
         ldmt = self.active_control.last_drops_minimum_time
         if ldmt:
             try:
@@ -413,6 +412,7 @@ class FlowSequencer (Singleton, I_TargetSetter):
                                   timeout=ldmt)
             except asyncio.exceptions.TimeoutError:
                 pass
+        await self._gate_flow_state_exit.wait()
         await self._end_sequence()
 
     async def _end_sequence(self):
@@ -485,27 +485,27 @@ class FlowSequencer (Singleton, I_TargetSetter):
         if self.active_state not in self.autotare_states:
             scale.hold_at_tare = False
             logger.debug(f"Scale: release - {self.active_state.name}")
-            await send_to_outbound_queue(AutoTareNotification(
+            await send_to_outbound_pipes(AutoTareNotification(
                 AutoTareNotificationAction.DISABLED))
             return
         try:
             await self._gate_sequence_start.wait()
             scale.hold_at_tare = True
             logger.debug("Scale: hold at tare")
-            await send_to_outbound_queue(AutoTareNotification(
+            await send_to_outbound_pipes(AutoTareNotification(
                 AutoTareNotificationAction.ENABLED))
 
             await self._gate_expect_drops.wait()
             scale.hold_at_tare = False
             logger.debug("Scale: release")
-            await send_to_outbound_queue(AutoTareNotification(
+            await send_to_outbound_pipes(AutoTareNotification(
                 AutoTareNotificationAction.DISABLED))
 
         except asyncio.CancelledError:
             scale = self._scale_processor.scale
             scale.hold_at_tare = False
             logger.info("Scale: release - on cancel")
-            await send_to_outbound_queue(AutoTareNotification(
+            await send_to_outbound_pipes(AutoTareNotification(
                 AutoTareNotificationAction.DISABLED))
             raise
 
@@ -642,18 +642,34 @@ class FlowSequencer (Singleton, I_TargetSetter):
             raise
 
     async def _sequence_recorder(self):
+        # Always enable recording, let the recorder decide
+
+        warnings.warn(
+            "de1._recorder_active will be removed shortly "
+            "in favor of database recording in another process.")
         de1 = self._de1
+
         try:
-            # Always enable recording, let the recorder decide
+            await pyDE1.database.write_notifications.create_history_record(self)
             await self._gate_sequence_start.wait()
+
+            FlowSequencer.database_queue.put_nowait(
+                pyDE1.database.write_notifications.RecorderControl(
+                    recording = True,
+                    sequence_id=SequencerGateNotification.sequence_id))
             de1._recorder_active = True
             logger.debug("Recorder: enable")
 
             await self._gate_sequence_complete.wait()
+            FlowSequencer.database_queue.put_nowait(
+                pyDE1.database.write_notifications.RecorderControl(
+                    recording = False,
+                    sequence_id=SequencerGateNotification.sequence_id))
             de1._recorder_active = False
             logger.debug("Recorder: disable")
 
         except asyncio.CancelledError:
+            FlowSequencer.database_queue.put_nowait(False)
             de1._recorder_active = False
             logger.info("Recorder: disable - on cancel")
             raise
@@ -730,7 +746,7 @@ class FlowSequencer (Singleton, I_TargetSetter):
             target = self.active_control.stop_at_weight
         else:
             target = None
-        await send_to_outbound_queue(StopAtNotification(
+        await send_to_outbound_pipes(StopAtNotification(
             stop_at=stop_at,
             action=action,
             target_value=target,
@@ -808,4 +824,4 @@ class FlowSequencer (Singleton, I_TargetSetter):
         try:
             return bmc.profile_can_override_tank_temperature
         except AttributeError:
-            return True
\ No newline at end of file
+            return True
