@@ -7,32 +7,38 @@ SPDX-License-Identifier: GPL-3.0-only
 """
 
 import asyncio
-import atexit
-import enum
 import logging
-import re
 import time
-import traceback
+import gc
 
-from typing import Optional, Callable, Coroutine, Any, Union
+from typing import Optional, Callable, Coroutine, Union
 
-from bleak import BleakScanner, BleakClient, BleakError
-from bleak import BleakError, BleakClient, BleakScanner
-from bleak.backends.bluezdbus.client import BleakClientBlueZDBus
 from bleak.backends.device import BLEDevice
 from bleak.backends.scanner import AdvertisementData
 
-import pyDE1.default_logger
 from pyDE1.bleak_client_wrapper import BleakClientWrapped
-from pyDE1.de1.exceptions import DE1APIValueError
+from pyDE1.scanner import BleakScannerWrapped
+from pyDE1.exceptions import DE1APIValueError, DE1NoAddressError
 from pyDE1.dispatcher.resource import ConnectivityEnum
-from pyDE1.event_manager import EventPayload, SubscribedEvent
+from pyDE1.event_manager import SubscribedEvent
 from pyDE1.event_manager.events import ConnectivityState, ConnectivityChange
-from pyDE1.scale.events import ScaleWeightUpdate, ScaleButtonPress, \
-    ScaleTareSeen
+from pyDE1.scale.events import ScaleWeightUpdate, ScaleTareSeen
+from pyDE1.scanner import _registered_ble_prefixes, DiscoveredDevices
+from pyDE1.config.bluetooth import CONNECT_TIMEOUT
 
 logger = logging.getLogger('Scale')
 
+# TODO: Can/should this be related to the class?
+#       If so, how should subclasses respond?
+
+# Used for factory and for BLE detection and filtering
+_prefix_to_constructor = dict()
+_recognized_scale_prefixes = set()
+
+
+def recognized_scale_prefixes():
+    return _recognized_scale_prefixes.copy()
+
 
 class ScaleError(RuntimeError):
     def __init__(self, *args, **kwargs):
@@ -52,7 +58,7 @@ class ScaleNotConnectedError(ScaleError):
 class Scale:
 
     def __init__(self):
-        self._address = None
+        self._address_or_bledevice: Optional[Union[str, BLEDevice]] = None
         self._name: Optional[str] = None
 
         # These are often model-specific, override in subclass init
@@ -75,10 +81,26 @@ class Scale:
         self._event_weight_update: SubscribedEvent = SubscribedEvent(self)
         self._event_button_press: SubscribedEvent = SubscribedEvent(self)
         self._event_tare_seen: SubscribedEvent = SubscribedEvent(self)
+
+        self._ready = asyncio.Event()
+
+        asyncio.get_event_loop().create_task(
+            self._event_connectivity.publish(
+                ConnectivityChange(arrival_time=time.time(),
+                                   state=ConnectivityState.NOT_READY)))
+
         self._estimated_period = self._nominal_period
         self._last_weight_update_received = 0
         self._last_tare_request_sent = 0
 
+        # See Scale.decommission()
+        self._to_decommission = (
+            '_event_connectivity',
+            '_event_weight_update',
+            '_event_button_press',
+            '_event_tare_seen',
+        )
+
         # TODO: Think about how to manage "tare seen"
         #       Could use asyncio.Event(), but what are the states
         #       and how do you "release" if it never arrives?
@@ -96,16 +118,6 @@ class Scale:
         asyncio.get_event_loop().create_task(
             self._event_weight_update.subscribe(self._create_self_callback()))
 
-    @classmethod
-    def device_adv_is_recognized_by(cls, device: BLEDevice, adv: AdvertisementData):
-        """
-        Think about how to have this scan the subclasses
-        for the most-specific match.
-
-        Also think about how to manage the possibility of a collection
-        of scales, that might already be instantiated
-        """
-        raise NotImplementedError
 
     @property
     def type(self):
@@ -113,7 +125,7 @@ class Scale:
 
     @property
     def address(self):
-        addr = self._address
+        addr = self._address_or_bledevice
         if isinstance(addr, BLEDevice):
             addr = addr.address
         return addr
@@ -122,13 +134,11 @@ class Scale:
     def address(self, address: Union[BLEDevice, str]):
         if self.address is not None:
             raise DE1APIValueError(
-                "Changing the scale address is not yet supported")
-        # If using BLEDevice directly
-        # AttributeError: 'BleakClientBlueZDBus' object has no attribute 'lower'
-        if isinstance(address, BleakClientBlueZDBus):
-            self._address = address.address
-        else:
-            self._address = address
+                "Changing the Scale address is not yet supported")
+        self._address_or_bledevice = address
+        if isinstance(address, BLEDevice):
+            self._name = address.name
+        self._bleak_client = BleakClientWrapped(self._address_or_bledevice)
 
     @property
     def name(self):
@@ -138,37 +148,54 @@ class Scale:
     def sensor_lag(self):
         return self._sensor_lag
 
-    async def connect(self, timeout=5.0):
+    async def connect(self, timeout: Optional[float] = None):
+
+        if timeout is None:
+            timeout = CONNECT_TIMEOUT
+
         class_name = type(self).__name__
         logger.info(f"Connecting to {class_name} at {self.address}")
-        if self._bleak_client is None:
-            if self.address is None:
-                raise ScaleNoAddressError
-            self._bleak_client = BleakClientWrapped(self.address)
+
+        assert self._bleak_client is not None
         self._bleak_client.set_disconnected_callback(
             self._create_disconnect_callback()
         )
+
         await asyncio.gather(self._event_connectivity.publish(
             ConnectivityChange(arrival_time=time.time(),
                                state=ConnectivityState.CONNECTING)),
             self._bleak_client.connect(timeout=timeout),
         )
+
         if self.is_connected:
-            self._address = self._bleak_client.address
+            self._address_or_bledevice = self._bleak_client.address
             if self.name is None:
                 self._name = self._bleak_client.name
             logger.info(f"Connected to {class_name} at {self.address}")
             await self._event_connectivity.publish(
                 ConnectivityChange(arrival_time=time.time(),
                                    state=ConnectivityState.CONNECTED))
+            # This can take some time, potentially delaying DE1 connection
+            # At least BlueZ doesn't like concurrent connection requests
+            asyncio.create_task(self.standard_initialization())
+
+            # TODO: Does the ScaleProcessor get properly reset?
+
         else:
             logger.error(
                 f"Connection failed to {class_name} at {self.address}")
-            await self._event_connectivity.publish(
-                ConnectivityChange(arrival_time=time.time(),
-                                   state=ConnectivityState.DISCONNECTED))
+            await asyncio.gather(self._notify_not_ready,
+                                 self._event_connectivity.publish(
+                                     ConnectivityChange(
+                                         arrival_time=time.time(),
+                                         state=ConnectivityState.DISCONNECTED))
+                                 )
 
     async def standard_initialization(self, hold_notification=False):
+        """
+        :param hold_notification: Since subclass may need to do more
+        """
+        logger.info("Scale.standard_initialization()")
         await self.display_on()
         await self.start_sending_weight_updates()
         if self.supports_button_press:
@@ -177,11 +204,14 @@ class Scale:
             await self._notify_ready()
 
     async def _notify_ready(self):
+        self._ready.set()
         await self._event_connectivity.publish(
             ConnectivityChange(arrival_time=time.time(),
                                state=ConnectivityState.READY))
+        logger.info("Ready")
 
     async def _notify_not_ready(self):
+        self._ready.clear()
         await self._event_connectivity.publish(
             ConnectivityChange(arrival_time=time.time(),
                                state=ConnectivityState.NOT_READY))
@@ -190,13 +220,18 @@ class Scale:
         class_name = type(self).__name__
         logger.info(f"Disconnecting from {class_name}")
         if self._bleak_client is None:
-            logger.info(f"Disconnecting from {class_name}; no client")
-        await asyncio.gather(
-            self._bleak_client.disconnect(),
-            self._event_connectivity.publish(
-                ConnectivityChange(arrival_time=time.time(),
-                                   state=ConnectivityState.DISCONNECTING))
-        )
+            logger.info(f"Disconnecting from {class_name}, no client")
+            return
+        if self.is_connected:
+            logger.debug(f"await gather disconnect, not ready, disconnecting")
+            gr = await asyncio.gather(
+                self._bleak_client.disconnect(),
+                self._notify_not_ready(),
+                self._event_connectivity.publish(
+                    ConnectivityChange(arrival_time=time.time(),
+                                       state=ConnectivityState.DISCONNECTING))
+            )
+
         if self.is_connected:
             logger.error(
                 f"Disconnect failed from {class_name} at {self.address}")
@@ -204,7 +239,9 @@ class Scale:
                 ConnectivityChange(arrival_time=time.time(),
                                    state=ConnectivityState.CONNECTED))
         else:
-            logger.info(f"Scale.disconnect(): Disconnected from {class_name} at {self.address}")
+            logger.info(
+                f"Scale.disconnect(): Disconnected from {class_name} "
+                f"at {self.address}")
             await self._event_connectivity.publish(
                 ConnectivityChange(arrival_time=time.time(),
                                    state=ConnectivityState.DISCONNECTED))
@@ -276,7 +313,7 @@ class Scale:
         if do_it:
             await self.tare()
 
-    async def display_bool(self, on):
+    async def display_bool(self, on: bool):
         if on:
             await self.display_on()
         else:
@@ -349,14 +386,32 @@ class Scale:
 
         return self_callback
 
+    def decommission(self):
+        # A Scale has several self-references that may prevent GC
+        # Rather than try to deal with weakref, just remove the callbacks
+        # Oh, and the list of those collbacks!
+        logger.info(f'Decommissioning {self} at {self.address}')
+
+        self._bleak_client.set_disconnected_callback(None)
+        for break_ref in self._to_decommission:
+            setattr(self, break_ref, None)
+        self._to_decommission = None
+
+        # logger.debug("Before GC")
+        # for ref in gc.get_referrers(self):
+        #     logger.info(f"0x{id(self):x} <== {ref}")
+        #
+        # logger.info(f"0x{id(self):x} is_finalized: {gc.is_finalized(self)}")
+
     # TODO: Decide how to handle  self._disconnected_callback
     #   disconnected_callback (callable): Callback that will be scheduled in the
     #   event loop when the client is disconnected. The callable must take one
     #   argument, which will be this client object.
 
     # The callback seems to be expected to be a "plain" function
-    # RuntimeWarning: coroutine 'DE1._create_disconnect_callback.<locals>.disconnect_callback'
-    #                 was never awaited
+    #     task.add_done_callback(
+    #         lambda _: self._disconnected_callback(self)
+    #     )
 
     def _create_disconnect_callback(self) -> Callable:
         scale = self
@@ -369,9 +424,12 @@ class Scale:
                 f"Disconnected from {class_name} at {client.address}, "
                 "willful_disconnect: "
                 f"{client.willful_disconnect}")
-            asyncio.create_task(scale._event_connectivity.publish(
+            asyncio.ensure_future(asyncio.gather(
+                self._notify_not_ready(),
+                scale._event_connectivity.publish(
                 ConnectivityChange(arrival_time=time.time(),
                                    state=ConnectivityState.DISCONNECTED)))
+            )
 
         return disconnect_callback
 
@@ -439,20 +497,49 @@ class Scale:
     # For API
     @property
     def connectivity(self):
+        retval = ConnectivityEnum.NOT_CONNECTED
         if self.is_connected:
-            retval = ConnectivityEnum.CONNECTED
-        else:
-            # intentionally vague, as PUT "connecting" isn't needed
-            retval = ConnectivityEnum.NOT_CONNECTED
+            if self._ready.is_set():
+                retval = ConnectivityEnum.READY
+            else:
+                retval = ConnectivityEnum.CONNECTED
         return retval
 
-    async def connectivity_setter(self, value: ConnectivityEnum):
+    async def connectivity_setter(self, value):
         assert isinstance(value, ConnectivityEnum), \
-            f"mode of {value} not a ConnectivityEnum " \
-            "in Scale.connectivity_setter()"
-        if value is ConnectivityEnum.CONNECTED \
-                and not self.is_connected:
+            f"mode of {value} not a ConnectivityEnum"
+        if value is ConnectivityEnum.CONNECTED and not self.is_connected:
             await self.connect()
-        elif value is ConnectivityEnum.NOT_CONNECTED \
-                and self.is_connected:
-            await self.disconnect()
\ No newline at end of file
+        elif value is ConnectivityEnum.NOT_CONNECTED and self.is_connected:
+            await self.disconnect()
+        else:
+            raise DE1APIValueError(
+                "Only CONNECTED and NOT_CONNECTED can be set, "
+                f"not {value}")
+
+    @staticmethod
+    def register_constructor(constructor: Callable, prefix: str):
+        _prefix_to_constructor[prefix] = constructor
+        _recognized_scale_prefixes.add(prefix)
+        _registered_ble_prefixes.add(prefix)
+
+
+def scale_factory(ble_device: BLEDevice)-> Scale:
+    constructor = None
+    try:
+        constructor = _prefix_to_constructor[ble_device.name]
+    except KeyError:
+        for prefix in _prefix_to_constructor.values():
+            if ble_device.name.startswith(prefix):
+                constructor = _prefix_to_constructor[prefix]
+    if constructor is None:
+        raise DE1APIValueError(
+            f"No recognized scale registered for {ble_device.name}"
+        )
+    logger.debug(f"Creating a new instance of {constructor} "
+                 f"from {ble_device}")
+    scale: Scale = constructor()
+    scale.address = ble_device
+    return scale
+
+
